@@ -1,40 +1,46 @@
 /*
-    Copyright (C) 2000 Paul Davis
-
-    This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-
-*/
+ * Copyright (C) 2005-2018 Paul Davis <paul@linuxaudiosystems.com>
+ * Copyright (C) 2005 Taybin Rutkin <taybin@taybin.com>
+ * Copyright (C) 2006-2012 David Robillard <d@drobilla.net>
+ * Copyright (C) 2008 Hans Baier <hansfbaier@googlemail.com>
+ * Copyright (C) 2009-2012 Carl Hetherington <carl@carlh.net>
+ * Copyright (C) 2013-2019 Robin Gareus <robin@gareus.org>
+ * Copyright (C) 2015-2016 Tim Mayberry <mojofunk@gmail.com>
+ * Copyright (C) 2015-2017 Nick Mainsbridge <mainsbridge@gmail.com>
+ * Copyright (C) 2015-2019 Ben Loftis <ben@harrisonconsoles.com>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+ */
 
 #include <cstdlib>
 #include <cmath>
 #include <algorithm>
 #include <typeinfo>
 
-#include "pbd/stacktrace.h"
-
 #include "ardour/audio_track.h"
 #include "ardour/midi_track.h"
 #include "ardour/midi_region.h"
-#include "ardour/region_factory.h"
 #include "ardour/profile.h"
+#include "ardour/region_factory.h"
 
 #include "canvas/canvas.h"
 #include "canvas/text.h"
 #include "canvas/scroll_group.h"
 
 #include "editor.h"
+#include "editor_sections.h"
 #include "keyboard.h"
 #include "public_editor.h"
 #include "audio_region_view.h"
@@ -48,10 +54,11 @@
 #include "editor_drag.h"
 #include "midi_time_axis.h"
 #include "editor_regions.h"
+#include "editor_sources.h"
 #include "ui_config.h"
 #include "verbose_cursor.h"
 
-#include "i18n.h"
+#include "pbd/i18n.h"
 
 using namespace std;
 using namespace ARDOUR;
@@ -75,11 +82,7 @@ Editor::track_canvas_scroll (GdkEventScroll* ev)
 	switch (direction) {
 	case GDK_SCROLL_UP:
 		if (Keyboard::modifier_state_equals (ev->state, Keyboard::ScrollZoomHorizontalModifier)) {
-			if (UIConfiguration::instance().get_use_mouse_position_as_zoom_focus_on_scroll()) {
-				temporal_zoom_step_mouse_focus (false);
-			} else {
-				temporal_zoom_step (false);
-			}
+			temporal_zoom_step_mouse_focus (false);
 			return true;
 		} else if (Keyboard::modifier_state_equals (ev->state, Keyboard::ScrollHorizontalModifier)) {
 			scroll_left_step ();
@@ -104,11 +107,7 @@ Editor::track_canvas_scroll (GdkEventScroll* ev)
 
 	case GDK_SCROLL_DOWN:
 		if (Keyboard::modifier_state_equals (ev->state, Keyboard::ScrollZoomHorizontalModifier)) {
-			if (UIConfiguration::instance().get_use_mouse_position_as_zoom_focus_on_scroll()) {
-				temporal_zoom_step_mouse_focus (true);
-			} else {
-				temporal_zoom_step (true);
-			}
+			temporal_zoom_step_mouse_focus (true);
 			return true;
 		} else if (Keyboard::modifier_state_equals (ev->state, Keyboard::ScrollHorizontalModifier)) {
 			scroll_right_step ();
@@ -167,11 +166,11 @@ bool
 Editor::track_canvas_button_press_event (GdkEventButton *event)
 {
 	_track_canvas->grab_focus();
-	if (!Keyboard::is_context_menu_event (event)) {
+	if (!internal_editing() && !Keyboard::is_context_menu_event (event)) {
 		begin_reversible_selection_op (X_("Clear Selection Click (track canvas)"));
 		selection->clear ();
 		commit_reversible_selection_op();
-        }
+	}
 	return false;
 }
 
@@ -179,10 +178,18 @@ bool
 Editor::track_canvas_button_release_event (GdkEventButton *event)
 {
 	if (!Keyboard::is_context_menu_event (event)) {
-                if (_drags->active ()) {
-                        _drags->end_grab ((GdkEvent*) event);
-                }
-        }
+		if (_drags->active ()) {
+
+			GdkEvent copy = *((GdkEvent*) event);
+			Duple winpos = Duple (event->x, event->y);
+			Duple where = _track_canvas->window_to_canvas (winpos);
+
+			copy.button.x = where.x;
+			copy.button.y = where.y;
+
+			_drags->end_grab (&copy);
+		}
+	}
 	return false;
 }
 
@@ -198,6 +205,10 @@ Editor::track_canvas_motion_notify_event (GdkEventMotion */*event*/)
 bool
 Editor::typed_event (ArdourCanvas::Item* item, GdkEvent *event, ItemType type)
 {
+	if (!session () || session()->loading () || session()->deletion_in_progress ()) {
+		return false;
+	}
+
 	gint ret = FALSE;
 
 	switch (event->type) {
@@ -364,6 +375,13 @@ Editor::canvas_stream_view_event (GdkEvent *event, ArdourCanvas::Item* item, Rou
 bool
 Editor::canvas_automation_track_event (GdkEvent *event, ArdourCanvas::Item* item, AutomationTimeAxisView *atv)
 {
+	if (atv->parameter().type() == MidiVelocityAutomation) {
+		/* no event handling for velocity tracks until we can make the
+		   automation control affect note velocity.
+		*/
+		return false;
+	}
+
 	bool ret = false;
 
 	switch (event->type) {
@@ -393,6 +411,18 @@ Editor::canvas_automation_track_event (GdkEvent *event, ArdourCanvas::Item* item
 		ret = leave_handler (item, event, AutomationTrackItem);
 		break;
 
+	case GDK_KEY_PRESS:
+		if (_drags->active()) {
+			return _drags->mid_drag_key_event (&event->key);
+		}
+		break;
+
+	case GDK_KEY_RELEASE:
+		if (_drags->active()) {
+			return _drags->mid_drag_key_event (&event->key);
+		}
+		break;
+
 	default:
 		break;
 	}
@@ -532,7 +562,6 @@ Editor::canvas_fade_in_handle_event (GdkEvent *event, ArdourCanvas::Item* item,
 
 	case GDK_BUTTON_RELEASE:
 		ret = button_release_handler (item, event, trim ? FadeInTrimHandleItem : FadeInHandleItem);
-		maybe_locate_with_edit_preroll ( rv->region()->position() );
 		break;
 
 	case GDK_MOTION_NOTIFY:
@@ -616,7 +645,6 @@ Editor::canvas_fade_out_handle_event (GdkEvent *event, ArdourCanvas::Item* item,
 
 	case GDK_BUTTON_RELEASE:
 		ret = button_release_handler (item, event, trim ? FadeOutTrimHandleItem : FadeOutHandleItem);
-		maybe_locate_with_edit_preroll ( rv->region()->last_frame() - rv->get_fade_out_shape_width() );
 		break;
 
 	case GDK_MOTION_NOTIFY:
@@ -639,9 +667,9 @@ Editor::canvas_fade_out_handle_event (GdkEvent *event, ArdourCanvas::Item* item,
 }
 
 struct DescendingRegionLayerSorter {
-    bool operator()(boost::shared_ptr<Region> a, boost::shared_ptr<Region> b) {
-	    return a->layer() > b->layer();
-    }
+	bool operator()(std::shared_ptr<Region> a, std::shared_ptr<Region> b) {
+		return a->layer() > b->layer();
+	}
 };
 
 bool
@@ -656,13 +684,6 @@ Editor::canvas_control_point_event (GdkEvent *event, ArdourCanvas::Item* item, C
 		clicked_routeview = dynamic_cast<RouteTimeAxisView*>(clicked_axisview);
 		clicked_regionview = 0;
 		break;
-
-	case GDK_SCROLL_UP:
-		break;
-
-	case GDK_SCROLL_DOWN:
-		break;
-
 	default:
 		break;
 	}
@@ -670,17 +691,33 @@ Editor::canvas_control_point_event (GdkEvent *event, ArdourCanvas::Item* item, C
 	return typed_event (item, event, ControlPointItem);
 }
 
+bool
+Editor::canvas_velocity_event (GdkEvent *event, ArdourCanvas::Item* item)
+{
+	return typed_event (item, event, VelocityItem);
+}
+
+bool
+Editor::canvas_velocity_base_event (GdkEvent *event, ArdourCanvas::Item* item)
+{
+	return typed_event (item, event, VelocityBaseItem);
+}
+
 bool
 Editor::canvas_line_event (GdkEvent *event, ArdourCanvas::Item* item, AutomationLine* al)
 {
 	ItemType type;
-	AudioRegionGainLine* gl;
-	if ((gl = dynamic_cast<AudioRegionGainLine*> (al)) != 0) {
+	RegionFxLine* rfl;
+	if ((rfl = dynamic_cast<RegionFxLine*> (al)) != 0) {
 		type = GainLineItem;
-		clicked_regionview = &gl->region_view ();
+		if (event->type == GDK_BUTTON_PRESS) {
+			clicked_regionview = &rfl->region_view ();
+		}
 	} else {
 		type = AutomationLineItem;
-		clicked_regionview = 0;
+		if (event->type == GDK_BUTTON_PRESS) {
+			clicked_regionview = 0;
+		}
 	}
 
 	clicked_control_point = 0;
@@ -871,7 +908,7 @@ Editor::canvas_region_view_name_highlight_event (GdkEvent* event, ArdourCanvas::
 		break;
 	case GDK_MOTION_NOTIFY:
 		motion_handler (item, event);
-		ret = true; // force this to avoid progagating the event into the regionview
+		ret = true; // force this to avoid propagating the event into the regionview
 		break;
 	case GDK_ENTER_NOTIFY:
 		ret = enter_handler (item, event, RegionViewNameHighlight);
@@ -974,27 +1011,9 @@ Editor::canvas_marker_event (GdkEvent *event, ArdourCanvas::Item* item, ArdourMa
 }
 
 bool
-Editor::canvas_marker_bar_event (GdkEvent *event, ArdourCanvas::Item* item)
+Editor::canvas_selection_marker_event (GdkEvent *event, ArdourCanvas::Item* item)
 {
-	return typed_event (item, event, MarkerBarItem);
-}
-
-bool
-Editor::canvas_range_marker_bar_event (GdkEvent *event, ArdourCanvas::Item* item)
-{
-	return typed_event (item, event, RangeMarkerBarItem);
-}
-
-bool
-Editor::canvas_transport_marker_bar_event (GdkEvent *event, ArdourCanvas::Item* item)
-{
-	return typed_event (item, event, TransportMarkerBarItem);
-}
-
-bool
-Editor::canvas_cd_marker_bar_event (GdkEvent *event, ArdourCanvas::Item* item)
-{
-	return typed_event (item, event, CdMarkerBarItem);
+	return typed_event (item, event, SelectionMarkerItem);
 }
 
 bool
@@ -1004,17 +1023,29 @@ Editor::canvas_videotl_bar_event (GdkEvent *event, ArdourCanvas::Item* item)
 }
 
 bool
-Editor::canvas_tempo_marker_event (GdkEvent *event, ArdourCanvas::Item* item, TempoMarker* /*marker*/)
+Editor::canvas_tempo_marker_event (GdkEvent *event, ArdourCanvas::Item* item, TempoMarker* marker)
 {
 	return typed_event (item, event, TempoMarkerItem);
 }
 
+bool
+Editor::canvas_tempo_curve_event (GdkEvent *event, ArdourCanvas::Item* item, TempoCurve* /*marker*/)
+{
+	return typed_event (item, event, TempoCurveItem);
+}
+
 bool
 Editor::canvas_meter_marker_event (GdkEvent *event, ArdourCanvas::Item* item, MeterMarker* /*marker*/)
 {
 	return typed_event (item, event, MeterMarkerItem);
 }
 
+bool
+Editor::canvas_bbt_marker_event (GdkEvent *event, ArdourCanvas::Item* item, BBTMarker* /*marker*/)
+{
+	return typed_event (item, event, BBTMarkerItem);
+}
+
 bool
 Editor::canvas_ruler_event (GdkEvent *event, ArdourCanvas::Item* item, ItemType type)
 {
@@ -1028,11 +1059,12 @@ Editor::canvas_ruler_event (GdkEvent *event, ArdourCanvas::Item* item, ItemType
 
 		switch (event->scroll.direction) {
 		case GDK_SCROLL_UP:
-			if (Keyboard::modifier_state_equals(event->scroll.state,
-			                                    Keyboard::ScrollHorizontalModifier)) {
+			if (Keyboard::modifier_state_equals (event->scroll.state, Keyboard::ScrollHorizontalModifier)) {
 				scroll_left_step ();
 			} else if (UIConfiguration::instance().get_use_mouse_position_as_zoom_focus_on_scroll()) {
 				temporal_zoom_step_mouse_focus (false);
+			} else if (Keyboard::modifier_state_equals (event->scroll.state, Keyboard::PrimaryModifier)) {
+				temporal_zoom_step_mouse_focus (false);
 			} else {
 				temporal_zoom_step (false);
 			}
@@ -1040,11 +1072,12 @@ Editor::canvas_ruler_event (GdkEvent *event, ArdourCanvas::Item* item, ItemType
 			break;
 
 		case GDK_SCROLL_DOWN:
-			if (Keyboard::modifier_state_equals(event->scroll.state,
-			                                    Keyboard::ScrollHorizontalModifier)) {
+			if (Keyboard::modifier_state_equals (event->scroll.state, Keyboard::ScrollHorizontalModifier)) {
 				scroll_right_step ();
 			} else if (UIConfiguration::instance().get_use_mouse_position_as_zoom_focus_on_scroll()) {
 				temporal_zoom_step_mouse_focus (true);
+			} else if (Keyboard::modifier_state_equals (event->scroll.state, Keyboard::PrimaryModifier)) {
+				temporal_zoom_step_mouse_focus (true);
 			} else {
 				temporal_zoom_step (true);
 			}
@@ -1072,15 +1105,96 @@ Editor::canvas_ruler_event (GdkEvent *event, ArdourCanvas::Item* item, ItemType
 }
 
 bool
-Editor::canvas_tempo_bar_event (GdkEvent *event, ArdourCanvas::Item* item)
+Editor::canvas_ruler_bar_event (GdkEvent *event, ArdourCanvas::Item* item, ItemType type, std::string const& color_name)
 {
-	return typed_event (item, event, TempoBarItem);
+	/* XXX consolidate with Editor::canvas_ruler_event ? */
+
+	switch (event->type) {
+	case GDK_ENTER_NOTIFY:
+		if (event->crossing.detail != GDK_NOTIFY_INFERIOR) {
+			item->set_fill_color (Gtkmm2ext::HSV(UIConfiguration::instance().color_mod (color_name, "marker bar")).lighter(0.20).color());
+		}
+		break;
+	case GDK_LEAVE_NOTIFY:
+		if (event->crossing.detail != GDK_NOTIFY_INFERIOR) {
+			item->set_fill_color (UIConfiguration::instance().color_mod (color_name, "marker bar"));
+		}
+		break;
+	default:
+		break;
+	};
+
+	return typed_event (item, event, type);
 }
 
 bool
-Editor::canvas_meter_bar_event (GdkEvent *event, ArdourCanvas::Item* item)
+Editor::section_rect_event (GdkEvent* ev, Location* loc, ArdourCanvas::Rectangle* rect, std::string color)
 {
-	return typed_event (item, event, MeterBarItem);
+	switch (ev->type) {
+		case GDK_ENTER_NOTIFY:
+			if (UIConfiguration::instance ().get_widget_prelight ()) {
+				rect->set_fill_color (Gtkmm2ext::HSV (UIConfiguration::instance().color (color)).lighter (0.25));
+				return true;
+			}
+			break;
+		case GDK_LEAVE_NOTIFY:
+			if (UIConfiguration::instance ().get_widget_prelight ()) {
+				rect->set_fill_color (UIConfiguration::instance().color (color));
+				return true;
+			}
+			break;
+		case GDK_BUTTON_PRESS:
+			if (Keyboard::modifier_state_equals (ev->button.state, Keyboard::PrimaryModifier)) {
+				/* used to add markers */
+				return false;
+			}
+			if (ev->button.button == 1) {
+				_session->request_locate (loc->start().samples());
+			}
+			return true;
+		case GDK_2BUTTON_PRESS:
+		case GDK_3BUTTON_PRESS:
+			if (Keyboard::modifier_state_equals (ev->button.state, Keyboard::PrimaryModifier)) {
+				return false;
+			}
+			if (ev->button.button == 1) {
+				assert (find_location_markers (loc));
+				rename_marker (find_location_markers (loc)->start);
+				return true;
+			}
+			break;
+		case GDK_BUTTON_RELEASE:
+			if (Keyboard::is_context_menu_event (&ev->button)) {
+				using namespace Menu_Helpers;
+
+				/* find section */
+				timepos_t start (loc->start ());
+				timepos_t end;
+				Location* l = _session->locations()->section_at (start, start, end);
+				assert (l);
+
+				timepos_t where (canvas_event_time (ev));
+				snap_to (where, Temporal::RoundNearest);
+
+				section_box_menu.set_name ("ArdourContextMenu");
+				MenuList& items (section_box_menu.items());
+				items.clear ();
+
+				items.push_back (MenuElem (_("New Arrangement Marker"), sigc::bind (sigc::mem_fun(*this, &Editor::add_location_mark_with_flag), where, Location::Flags(Location::IsMark | Location::IsSection), 0)));
+				items.push_back (MenuElem (_("Select Arrangement Section"), sigc::bind (sigc::mem_fun(*_sections, &EditorSections::select), l)));
+#if 0
+				items.push_back (SeparatorElem());
+				add_section_context_items (items);   //TODO: section_context_items needs to be modified to operate on the marker you clicked on, not the range selection (which might not exist)
+#endif
+
+				section_box_menu.popup (ev->button.button, ev->button.time);
+				return true;
+			}
+			break;
+		default:
+			break;
+	}
+	return false;
 }
 
 bool
@@ -1089,6 +1203,36 @@ Editor::canvas_playhead_cursor_event (GdkEvent *event, ArdourCanvas::Item* item)
 	return typed_event (item, event, PlayheadCursorItem);
 }
 
+bool
+Editor::canvas_section_box_event (GdkEvent *event)
+{
+	switch (event->type) {
+		case GDK_BUTTON_PRESS:
+			if (!Keyboard::modifier_state_equals (event->button.state, Keyboard::PrimaryModifier)
+			   && event->button.button == 1) {
+				_drags->set (new CursorDrag (this, *_playhead_cursor, false), event);
+			}
+			/*fallthrough*/
+		case GDK_2BUTTON_PRESS:
+			/*fallthrough*/
+		case GDK_3BUTTON_PRESS:
+			return !Keyboard::modifier_state_equals (event->button.state, Keyboard::PrimaryModifier);
+		case GDK_BUTTON_RELEASE:
+			if (Keyboard::is_context_menu_event (&event->button)) {
+				section_box_menu.set_name ("ArdourContextMenu");
+				Gtk::Menu_Helpers::MenuList& items (section_box_menu.items());
+				items.clear ();
+				add_section_context_items (items);
+				section_box_menu.popup (event->button.button, event->button.time);
+				return true;
+			}
+			return false;
+		default:
+			break;
+	}
+	return false;
+}
+
 bool
 Editor::canvas_note_event (GdkEvent *event, ArdourCanvas::Item* item)
 {
@@ -1117,8 +1261,8 @@ Editor::canvas_drop_zone_event (GdkEvent* event)
 
 	case GDK_SCROLL:
 		/* convert coordinates back into window space so that
-		   we can just call canvas_scroll_event().
-		*/
+		 * we can just call canvas_scroll_event().
+		 */
 		winpos = _track_canvas->canvas_to_window (Duple (event->scroll.x, event->scroll.y));
 		scroll = event->scroll;
 		scroll.x = winpos.x;
@@ -1132,6 +1276,10 @@ Editor::canvas_drop_zone_event (GdkEvent* event)
 	case GDK_LEAVE_NOTIFY:
 		return typed_event (_canvas_drop_zone, event, DropZoneItem);
 
+	case GDK_MOTION_NOTIFY:
+		return motion_handler (_canvas_drop_zone, event);
+		break;
+
 	default:
 		break;
 	}
@@ -1139,11 +1287,58 @@ Editor::canvas_drop_zone_event (GdkEvent* event)
 	return true;
 }
 
+bool
+Editor::canvas_grid_zone_event (GdkEvent* event)
+{
+	GdkEventScroll scroll;
+	ArdourCanvas::Duple winpos;
+
+
+	switch (event->type) {
+
+		case GDK_BUTTON_PRESS:
+			if (event->button.button == 1) {
+				choose_mapping_drag (_canvas_grid_zone, event);
+			}
+			break;
+
+		case GDK_BUTTON_RELEASE:
+			return typed_event (_canvas_grid_zone, event, GridZoneItem);
+			break;
+
+		case GDK_SCROLL:
+			/* convert coordinates back into window space so that
+			 * we can just call canvas_scroll_event().
+			 */
+			winpos   = _track_canvas->canvas_to_window (Duple (event->scroll.x, event->scroll.y));
+			scroll   = event->scroll;
+			scroll.x = winpos.x;
+			scroll.y = winpos.y;
+			return canvas_scroll_event (&scroll, true);
+			break;
+
+		case GDK_ENTER_NOTIFY:
+			return typed_event (_canvas_grid_zone, event, GridZoneItem);
+
+		case GDK_LEAVE_NOTIFY:
+			return typed_event (_canvas_grid_zone, event, GridZoneItem);
+
+		case GDK_MOTION_NOTIFY:
+			return motion_handler (_canvas_grid_zone, event);
+			break;
+
+		default:
+			break;
+	}
+
+	return true;
+}
+
 bool
 Editor::track_canvas_drag_motion (Glib::RefPtr<Gdk::DragContext> const& context, int x, int y, guint time)
 {
-	boost::shared_ptr<Region> region;
-	boost::shared_ptr<Region> region_copy;
+	std::shared_ptr<Region> region;
+	std::shared_ptr<Region> region_copy;
 	RouteTimeAxisView* rtav;
 	GdkEvent event;
 	double px;
@@ -1184,33 +1379,18 @@ Editor::track_canvas_drag_motion (Glib::RefPtr<Gdk::DragContext> const& context,
 	}
 
 	if (can_drop) {
-		region = _regions->get_dragged_region ();
 
-		if (region) {
-
-			if (tv.first == 0
-			    && (
-			        boost::dynamic_pointer_cast<AudioRegion> (region) != 0 ||
-			        boost::dynamic_pointer_cast<MidiRegion> (region) != 0
-			       )
-			   )
-			{
+		if (target == "x-ardour/region.pbdid") {
+			if (tv.first == 0 && pbdid_dragged_dt != DataType::NIL) {
 				/* drop to drop-zone */
-				context->drag_status (context->get_suggested_action(), time);
+				context->drag_status (Gdk::ACTION_COPY, time);
 				return true;
 			}
 
-			if ((boost::dynamic_pointer_cast<AudioRegion> (region) != 0 &&
-			     dynamic_cast<AudioTimeAxisView*> (tv.first) != 0) ||
-			    (boost::dynamic_pointer_cast<MidiRegion> (region) != 0 &&
-			     dynamic_cast<MidiTimeAxisView*> (tv.first) != 0)) {
-
-				/* audio to audio
-				   OR
-				   midi to midi
-				*/
-
-				context->drag_status (context->get_suggested_action(), time);
+			if ((pbdid_dragged_dt == DataType::AUDIO && dynamic_cast<AudioTimeAxisView*> (tv.first) != 0) ||
+			    (pbdid_dragged_dt == DataType::MIDI  && dynamic_cast<MidiTimeAxisView*> (tv.first) != 0)) {
+				/* audio to audio OR MIDI to MIDI */
+				context->drag_status (Gdk::ACTION_COPY, time);
 				return true;
 			}
 		} else {
@@ -1219,7 +1399,7 @@ Editor::track_canvas_drag_motion (Glib::RefPtr<Gdk::DragContext> const& context,
 			 * TODO: check if file is audio/midi, allow drops on same track-type only,
 			 * currently: if audio is dropped on a midi-track, it is only added to the region-list
 			 */
-			if (Profile->get_sae() || UIConfiguration::instance().get_only_copy_imported_files()) {
+			if (UIConfiguration::instance().get_only_copy_imported_files()) {
 				context->drag_status(Gdk::ACTION_COPY, time);
 			} else {
 				if ((context->get_actions() & (Gdk::ACTION_COPY | Gdk::ACTION_LINK | Gdk::ACTION_MOVE)) == Gdk::ACTION_COPY) {
@@ -1239,9 +1419,9 @@ Editor::track_canvas_drag_motion (Glib::RefPtr<Gdk::DragContext> const& context,
 
 void
 Editor::drop_regions (const Glib::RefPtr<Gdk::DragContext>& /*context*/,
-		      int x, int y,
-		      const SelectionData& /*data*/,
-		      guint /*info*/, guint /*time*/)
+                      int x, int y,
+                      const SelectionData& data,
+                      guint /*info*/, guint /*time*/)
 {
 	GdkEvent event;
 	double px;
@@ -1252,9 +1432,11 @@ Editor::drop_regions (const Glib::RefPtr<Gdk::DragContext>& /*context*/,
 	event.button.y = y;
 	/* assume we're dragging with button 1 */
 	event.motion.state = Gdk::BUTTON1_MASK;
-	framepos_t const pos = window_event_sample (&event, &px, &py);
+	samplepos_t const pos = window_event_sample (&event, &px, &py);
+
+	PBD::ID rid (data.get_data_as_string ());
+	std::shared_ptr<Region> region = RegionFactory::region_by_id (rid);
 
-	boost::shared_ptr<Region> region = _regions->get_dragged_region ();
 	if (!region) { return; }
 
 	RouteTimeAxisView* rtav = 0;
@@ -1264,19 +1446,23 @@ Editor::drop_regions (const Glib::RefPtr<Gdk::DragContext>& /*context*/,
 		rtav = dynamic_cast<RouteTimeAxisView*> (tv.first);
 	} else {
 		try {
-			if (boost::dynamic_pointer_cast<AudioRegion> (region)) {
-				uint32_t output_chan = region->n_channels();
+			if (std::dynamic_pointer_cast<AudioRegion> (region)) {
+				uint32_t output_chan = region->sources().size();
 				if ((Config->get_output_auto_connect() & AutoConnectMaster) && session()->master_out()) {
 					output_chan =  session()->master_out()->n_inputs().n_audio();
 				}
-				list<boost::shared_ptr<AudioTrack> > audio_tracks;
-				audio_tracks = session()->new_audio_track (region->n_channels(), output_chan, ARDOUR::Normal, 0, 1, region->name());
-				rtav = axis_view_from_route (audio_tracks.front());
-			} else if (boost::dynamic_pointer_cast<MidiRegion> (region)) {
+				list<std::shared_ptr<AudioTrack> > audio_tracks;
+				audio_tracks = session()->new_audio_track (region->sources().size(), output_chan, 0, 1, region->name(), PresentationInfo::max_order);
+				rtav = dynamic_cast<RouteTimeAxisView*> (time_axis_view_from_stripable (audio_tracks.front()));
+			} else if (std::dynamic_pointer_cast<MidiRegion> (region)) {
 				ChanCount one_midi_port (DataType::MIDI, 1);
-				list<boost::shared_ptr<MidiTrack> > midi_tracks;
-				midi_tracks = session()->new_midi_track (one_midi_port, one_midi_port, boost::shared_ptr<ARDOUR::PluginInfo>(), ARDOUR::Normal, 0, 1, region->name());
-				rtav = axis_view_from_route (midi_tracks.front());
+				list<std::shared_ptr<MidiTrack> > midi_tracks;
+				midi_tracks = session()->new_midi_track (one_midi_port, one_midi_port,
+				                                         Config->get_strict_io () || Profile->get_mixbus (),
+				                                         std::shared_ptr<ARDOUR::PluginInfo>(),
+				                                         (ARDOUR::Plugin::PresetRecord*) 0,
+				                                         (ARDOUR::RouteGroup*) 0, 1, region->name(), PresentationInfo::max_order, Normal, true);
+				rtav = dynamic_cast<RouteTimeAxisView*> (time_axis_view_from_stripable (midi_tracks.front()));
 			} else {
 				return;
 			}
@@ -1287,12 +1473,12 @@ Editor::drop_regions (const Glib::RefPtr<Gdk::DragContext>& /*context*/,
 	}
 
 	if (rtav != 0 && rtav->is_track ()) {
-		boost::shared_ptr<Region> region_copy = RegionFactory::create (region, true);
+		std::shared_ptr<Region> region_copy = RegionFactory::create (region, true);
 
-		if ((boost::dynamic_pointer_cast<AudioRegion> (region_copy) != 0 && dynamic_cast<AudioTimeAxisView*> (rtav) != 0) ||
-		    (boost::dynamic_pointer_cast<MidiRegion> (region_copy) != 0 && dynamic_cast<MidiTimeAxisView*> (rtav) != 0)) {
-			_drags->set (new RegionInsertDrag (this, region_copy, rtav, pos), &event);
-			_drags->end_grab (0);
+		if ((std::dynamic_pointer_cast<AudioRegion> (region_copy) != 0 && dynamic_cast<AudioTimeAxisView*> (rtav) != 0) ||
+		    (std::dynamic_pointer_cast<MidiRegion> (region_copy) != 0 && dynamic_cast<MidiTimeAxisView*> (rtav) != 0)) {
+			_drags->set (new RegionInsertDrag (this, region_copy, rtav, timepos_t (pos), drag_time_domain (region_copy.get())), &event);
+			_drags->end_grab (&event);
 		}
 	}
 }
@@ -1338,4 +1524,3 @@ Editor::key_release_handler (ArdourCanvas::Item* item, GdkEvent* event, ItemType
 
 	return handled;
 }
-
