@@ -136,7 +136,7 @@ PortExportChannelSelector::fill_route_list ()
 	routes.sort (Stripable::Sorter ());
 
 	for (RouteList::iterator it = routes.begin(); it != routes.end(); ++it) {
-		if ((*it)->is_master () || (*it)->is_monitor ()) {
+		if ((*it)->is_main_bus ()) {
 			continue;
 		}
 		if ((*it)->output()->n_ports ().n_audio () > 0) {
@@ -733,8 +733,8 @@ TrackExportChannelSelector::sync_with_manager_state ()
 					PortExportChannel::PortSet route_ports;
 					PortExportChannel::PortSet intersection;
 
-					PortSet& ps (route->output()->ports ());
-					for (PortSet::audio_iterator p = ps.audio_begin (); p != ps.audio_end (); ++p) {
+					std::shared_ptr<PortSet> ps (route->output()->ports ());
+					for (PortSet::audio_iterator p = ps->audio_begin (); p != ps->audio_end (); ++p) {
 						route_ports.insert (*p);
 					}
 
@@ -838,7 +838,7 @@ TrackExportChannelSelector::fill_list()
 	for (RouteList::iterator it = routes.begin(); it != routes.end(); ++it) {
 		if (!std::dynamic_pointer_cast<Track>(*it)) {
 			// not a track, must be a bus
-			if ((*it)->is_master () || (*it)->is_monitor ()) {
+			if ((*it)->is_main_bus () && !(*it)->is_surround_master ()) {
 				continue;
 			}
 			if (!(*it)->active ()) {
