@@ -1,38 +1,55 @@
 /*
-    Copyright (C) 2008 Paul Davis
-    Author: Sakari Bergen
-
-    This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-
-*/
+ * Copyright (C) 2005-2006 Doug McLain <doug@nostar.net>
+ * Copyright (C) 2005-2006 Taybin Rutkin <taybin@taybin.com>
+ * Copyright (C) 2005-2008 Nick Mainsbridge <mainsbridge@gmail.com>
+ * Copyright (C) 2005-2017 Paul Davis <paul@linuxaudiosystems.com>
+ * Copyright (C) 2006-2011 David Robillard <d@drobilla.net>
+ * Copyright (C) 2007-2012 Tim Mayberry <mojofunk@gmail.com>
+ * Copyright (C) 2008-2013 Sakari Bergen <sakari.bergen@beatwaves.net>
+ * Copyright (C) 2009-2012 Carl Hetherington <carl@carlh.net>
+ * Copyright (C) 2013-2014 Colin Fletcher <colin.m.fletcher@googlemail.com>
+ * Copyright (C) 2015-2017 Robin Gareus <robin@gareus.org>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+ */
 
 
 #include <sigc++/signal.h>
 
 #include <gtkmm/messagedialog.h>
+#include <gtkmm/stock.h>
+
+#include "pbd/gstdio_compat.h"
+#include "pbd/file_utils.h"
 
 #include "ardour/audioregion.h"
+#include "ardour/export_channel_configuration.h"
+#include "ardour/export_format_specification.h"
+#include "ardour/export_format_manager.h"
 #include "ardour/export_status.h"
 #include "ardour/export_handler.h"
+#include "ardour/profile.h"
 
 #include "export_dialog.h"
 #include "export_report.h"
 #include "gui_thread.h"
+#include "mixer_ui.h"
 #include "nag.h"
+#include "ui_config.h"
 
-#include "i18n.h"
+#include "pbd/i18n.h"
 
 using namespace ARDOUR;
 using namespace PBD;
@@ -42,10 +59,12 @@ ExportDialog::ExportDialog (PublicEditor & editor, std::string title, ARDOUR::Ex
   : ArdourDialog (title)
   , type (type)
   , editor (editor)
-
-  , warn_label ("", Gtk::ALIGN_LEFT)
-  , list_files_label (_("<span color=\"#ffa755\">Some already existing files will be overwritten.</span>"), Gtk::ALIGN_RIGHT)
+  , warn_label ("", Gtk::ALIGN_START)
+  , list_files_label (_("<span color=\"#ffa755\">Some already existing files will be overwritten.</span>"), Gtk::ALIGN_END)
   , list_files_button (_("List files"))
+  , previous_progress (0)
+  , _initialized (false)
+  , _analysis_only (false)
 { }
 
 ExportDialog::~ExportDialog ()
@@ -80,7 +99,7 @@ ExportDialog::set_session (ARDOUR::Session* s)
 
 	TimeSelection const & time (editor.get_selection().time);
 	if (!time.empty()) {
-		profile_manager->set_selection_range (time.front().start, time.front().end);
+		profile_manager->set_selection_range (time.front().start().samples(), time.front().end().samples());
 	} else {
 		profile_manager->set_selection_range ();
 	}
@@ -95,9 +114,19 @@ ExportDialog::set_session (ARDOUR::Session* s)
 	preset_selector->CriticalSelectionChanged.connect (sigc::mem_fun (*this, &ExportDialog::sync_with_manager));
 	timespan_selector->CriticalSelectionChanged.connect (sigc::mem_fun (*this, &ExportDialog::update_warnings_and_example_filename));
 	channel_selector->CriticalSelectionChanged.connect (sigc::mem_fun (*this, &ExportDialog::update_warnings_and_example_filename));
+	channel_selector->CriticalSelectionChanged.connect (sigc::mem_fun (*this, &ExportDialog::update_realtime_selection));
 	file_notebook->CriticalSelectionChanged.connect (sigc::mem_fun (*this, &ExportDialog::update_warnings_and_example_filename));
 
-	update_warnings_and_example_filename ();
+	/* Catch major selection changes, and set the session dirty */
+
+	preset_selector->CriticalSelectionChanged.connect (sigc::mem_fun (*this, &ExportDialog::maybe_set_session_dirty));
+	timespan_selector->CriticalSelectionChanged.connect (sigc::mem_fun (*this, &ExportDialog::maybe_set_session_dirty));
+	channel_selector->CriticalSelectionChanged.connect (sigc::mem_fun (*this, &ExportDialog::maybe_set_session_dirty));
+	file_notebook->CriticalSelectionChanged.connect (sigc::mem_fun (*this, &ExportDialog::maybe_set_session_dirty));
+
+	_initialized = true;
+
+	_session->config.ParameterChanged.connect (*this, invalidator (*this), boost::bind (&ExportDialog::parameter_changed, this, _1), gui_context());
 }
 
 void
@@ -125,13 +154,15 @@ ExportDialog::init ()
 	progress_widget.pack_start (progress_bar, false, false, 6);
 
 	/* Buttons */
-
 	cancel_button = add_button (Gtk::Stock::CANCEL, RESPONSE_CANCEL);
+	analyze_button = add_button (_("Only Analyze"), RESPONSE_ANALYZE);
 	export_button = add_button (_("Export"), RESPONSE_FAST);
+
 	set_default_response (RESPONSE_FAST);
 
 	cancel_button->signal_clicked().connect (sigc::mem_fun (*this, &ExportDialog::close_dialog));
-	export_button->signal_clicked().connect (sigc::mem_fun (*this, &ExportDialog::do_export));
+	export_button->signal_clicked().connect (sigc::bind (sigc::mem_fun (*this, &ExportDialog::do_export), false));
+	analyze_button->signal_clicked().connect (sigc::bind (sigc::mem_fun (*this, &ExportDialog::do_export), true));
 
 	file_notebook->soundcloud_export_selector = soundcloud_selector;
 
@@ -146,7 +177,7 @@ ExportDialog::init_gui ()
 {
 	Gtk::Alignment * preset_align = Gtk::manage (new Gtk::Alignment());
 	preset_align->add (*preset_selector);
-	preset_align->set_padding (0, 12, 0, 0);
+	preset_align->set_padding (6, 8, 6, 6);
 
 	Gtk::VBox * file_format_selector = Gtk::manage (new Gtk::VBox());
 	file_format_selector->set_homogeneous (false);
@@ -204,6 +235,18 @@ ExportDialog::sync_with_manager ()
 	file_notebook->sync_with_manager ();
 
 	update_warnings_and_example_filename ();
+	update_realtime_selection ();
+}
+
+
+void
+ExportDialog::maybe_set_session_dirty ()
+{
+	/* Presumably after all initialization is finished, sync_with_manager means that something important changed. */
+	/* Let's prompt the user to save the session; otherwise these Export settings changes would be lost on re-open */
+	if (_initialized) {
+		_session->set_dirty();
+	}
 }
 
 void
@@ -218,10 +261,11 @@ ExportDialog::update_warnings_and_example_filename ()
 	list_files_string = "";
 
 	export_button->set_sensitive (true);
+	analyze_button->set_sensitive (true);
 
 	/* Add new warnings */
 
-	boost::shared_ptr<ExportProfileManager::Warnings> warnings = profile_manager->get_warnings();
+	std::shared_ptr<ExportProfileManager::Warnings> warnings = profile_manager->get_warnings();
 
 	for (std::list<string>::iterator it = warnings->errors.begin(); it != warnings->errors.end(); ++it) {
 		add_error (*it);
@@ -231,6 +275,11 @@ ExportDialog::update_warnings_and_example_filename ()
 		add_warning (*it);
 	}
 
+	/* add channel count warning */
+	if (channel_selector && channel_selector->channel_limit_reached ()) {
+		add_warning (_("A track or bus has more channels than the target."));
+	}
+
 	if (!warnings->conflicting_filenames.empty()) {
 		list_files_hbox.show ();
 		for (std::list<string>::iterator it = warnings->conflicting_filenames.begin(); it != warnings->conflicting_filenames.end(); ++it) {
@@ -244,12 +293,42 @@ ExportDialog::update_warnings_and_example_filename ()
 	file_notebook->update_example_filenames();
 }
 
+void
+ExportDialog::update_realtime_selection ()
+{
+	bool rt_ok = true;
+	switch (profile_manager->type ()) {
+		case ExportProfileManager::RegularExport:
+		case ExportProfileManager::RangeExport:
+		case ExportProfileManager::SelectionExport:
+			break;
+		case ExportProfileManager::RegionExport:
+			rt_ok = false;
+			break;
+		case ExportProfileManager::StemExport:
+			if (! static_cast<TrackExportChannelSelector*>(channel_selector.get())->track_output ()) {
+				rt_ok = false;
+			}
+			break;
+	}
+
+	timespan_selector->allow_realtime_export (rt_ok);
+}
+
+void
+ExportDialog::parameter_changed (std::string const& p)
+{
+	if (p == "realtime-export") {
+		update_realtime_selection ();
+	}
+}
+
 void
 ExportDialog::show_conflicting_files ()
 {
 	ArdourDialog dialog (_("Files that will be overwritten"), true);
 
-	Gtk::Label label ("", Gtk::ALIGN_LEFT);
+	Gtk::Label label ("", Gtk::ALIGN_START);
 	label.set_use_markup (true);
 	label.set_markup (list_files_string);
 
@@ -268,8 +347,18 @@ ExportDialog::soundcloud_upload_progress(double total, double now, std::string t
 }
 
 void
-ExportDialog::do_export ()
+ExportDialog::do_export (bool analysis_only)
 {
+	_analysis_only = analysis_only;
+	if (analysis_only) {
+		for (auto const& fmt : profile_manager->get_formats ()) {
+			std::shared_ptr<ExportFormatSpecification> fmp = fmt->format;
+			fmp->set_format_id (ExportFormatBase::F_None);
+			fmp->set_type (ExportFormatBase::T_None);
+			fmp->set_analyse (true);
+		}
+	}
+
 	try {
 		profile_manager->prepare_for_export ();
 		handler->soundcloud_username     = soundcloud_selector->username ();
@@ -289,6 +378,12 @@ ExportDialog::do_export ()
 				gui_context()
 				);
 #endif
+
+		_files_to_reimport.clear ();
+		Session::Exported.connect_same_thread (*this, sigc::bind (
+					[] (std::string, std::string fn, bool re, samplepos_t pos, ReImportMap* v) { if (re) { (*v)[pos].push_back (fn); } },
+					&_files_to_reimport));
+
 		handler->do_export ();
 		show_progress ();
 	} catch(std::exception & e) {
@@ -300,8 +395,11 @@ ExportDialog::do_export ()
 void
 ExportDialog::show_progress ()
 {
+	export_notebook.set_sensitive (false);
+
 	cancel_button->set_label (_("Stop Export"));
 	export_button->set_sensitive (false);
+	analyze_button->set_sensitive (false);
 
 	progress_bar.set_fraction (0.0);
 	warning_widget.hide_all();
@@ -319,26 +417,77 @@ ExportDialog::show_progress ()
 		}
 	}
 
-	status->finish ();
+	status->finish (TRS_UI);
+
+	if (!status->aborted() && !_files_to_reimport.empty ()) {
+		for (auto const& x : _files_to_reimport) {
+			timepos_t pos (x.first);
+			Editing::ImportDisposition disposition = Editing::ImportDistinctFiles;
+			editor.do_import (x.second, disposition, Editing::ImportAsTrack, SrcBest, SMFFileAndTrackName, SMFTempoIgnore, pos);
+		}
+	}
+
+	if (!status->aborted() && UIConfiguration::instance().get_save_export_mixer_screenshot ()) {
+		ExportProfileManager::TimespanStateList const& timespans = profile_manager->get_timespans();
+		ExportProfileManager::FilenameStateList const& filenames = profile_manager->get_filenames ();
+
+		std::list<std::string> paths;
+		for (ExportProfileManager::FilenameStateList::const_iterator fi = filenames.begin(); fi != filenames.end(); ++fi) {
+			for (ExportProfileManager::TimespanStateList::const_iterator ti = timespans.begin(); ti != timespans.end(); ++ti) {
+				ExportProfileManager::TimespanListPtr tlp = (*ti)->timespans;
+				for (ExportProfileManager::TimespanList::const_iterator eti = tlp->begin(); eti != tlp->end(); ++eti) {
+					(*fi)->filename->set_timespan (*eti);
+					paths.push_back ((*fi)->filename->get_path (ExportFormatSpecPtr ()) + "-mixer.png");
+				}
+			}
+		}
+
+		if (paths.size() > 0) {
+			PBD::info << string_compose(_("Writing Mixer Screenshot: %1."), paths.front()) << endmsg;
+			Mixer_UI::instance()->screenshot (paths.front());
+
+			std::list<std::string>::const_iterator it = paths.begin ();
+			++it;
+			for (; it != paths.end(); ++it) {
+				PBD::info << string_compose(_("Copying Mixer Screenshot: %1."), *it) << endmsg;
+				::g_unlink (it->c_str());
+				if (!hard_link (paths.front(), *it)) {
+					copy_file (paths.front(), *it);
+				}
+			}
+		}
+	}
+
+	if (!status->aborted() && _session->export_xruns () > 0) {
+		std::string txt = string_compose (_("There have been %1 dropouts during realtime-export."), _session->export_xruns ());
+		Gtk::MessageDialog msg (txt, false, Gtk::MESSAGE_WARNING, Gtk::BUTTONS_OK, true);
+		msg.run();
+	}
 
 	if (!status->aborted() && status->result_map.size() > 0) {
 		hide();
-		ExportReport er (_session, status);
-		er.run();
+		if (_analysis_only) {
+			ExportReport er (_("Export Report/Analysis"), status->result_map);
+			er.run();
+		} else {
+			ExportReport er (_session, status);
+			er.run();
+		}
 	}
 
 	if (!status->aborted()) {
 		hide();
-
-		NagScreen* ns = NagScreen::maybe_nag (_("export"));
-
-		if (ns) {
-			ns->nag ();
-			delete ns;
+		if (!ARDOUR::Profile->get_mixbus () && !ARDOUR::Profile->get_livetrax ()) {
+			NagScreen* ns = NagScreen::maybe_nag (_("export"));
+			if (ns) {
+				ns->nag ();
+				delete ns;
+			}
 		}
 	} else {
 		notify_errors ();
 	}
+	export_notebook.set_sensitive (true);
 }
 
 gint
@@ -346,28 +495,50 @@ ExportDialog::progress_timeout ()
 {
 	std::string status_text;
 	float progress = -1;
-	switch (status->active_job) {
-	case ExportStatus::Exporting:
-		status_text = string_compose (_("Exporting '%3' (timespan %1 of %2)"),
-		                              status->timespan, status->total_timespans, status->timespan_name);
-		progress = ((float) status->processed_frames_current_timespan) / status->total_frames_current_timespan;
-		break;
-	case ExportStatus::Normalizing:
-		status_text = string_compose (_("Normalizing '%3' (timespan %1 of %2)"),
-		                              status->timespan, status->total_timespans, status->timespan_name);
-		progress = ((float) status->current_normalize_cycle) / status->total_normalize_cycles;
-		break;
-	case ExportStatus::Tagging:
-		status_text = string_compose (_("Tagging '%3' (timespan %1 of %2)"),
-		                              status->timespan, status->total_timespans, status->timespan_name);
-		break;
-	case ExportStatus::Uploading:
-		status_text = string_compose (_("Uploading '%3' (timespan %1 of %2)"),
-		                              status->timespan, status->total_timespans, status->timespan_name);
-		break;
-	case ExportStatus::Command:
-		status_text = string_compose (_("Running Post Export Command for '%1'"), status->timespan_name);
-		break;
+
+	if (_analysis_only) {
+		switch (status->active_job) {
+			case ExportStatus::Exporting:
+				status_text = string_compose (_("Export for Analysis '%3' (timespan %1 of %2)"),
+				                              status->timespan, status->total_timespans, status->timespan_name);
+				progress = ((float) status->processed_samples_current_timespan) / status->total_samples_current_timespan;
+				break;
+			default:
+				status_text = string_compose (_("Analyzing '%3' (timespan %1 of %2)"),
+				                              status->timespan, status->total_timespans, status->timespan_name);
+				progress = ((float) status->current_postprocessing_cycle) / status->total_postprocessing_cycles;
+				break;
+		}
+	} else {
+
+		switch (status->active_job) {
+			case ExportStatus::Exporting:
+				status_text = string_compose (_("Exporting '%3' (timespan %1 of %2)"),
+				                              status->timespan, status->total_timespans, status->timespan_name);
+				progress = ((float) status->processed_samples_current_timespan) / status->total_samples_current_timespan;
+				break;
+			case ExportStatus::Normalizing:
+				status_text = string_compose (_("Normalizing '%3' (timespan %1 of %2)"),
+				                              status->timespan, status->total_timespans, status->timespan_name);
+				progress = ((float) status->current_postprocessing_cycle) / status->total_postprocessing_cycles;
+				break;
+			case ExportStatus::Encoding:
+				status_text = string_compose (_("Encoding '%3' (timespan %1 of %2)"),
+				                              status->timespan, status->total_timespans, status->timespan_name);
+				progress = ((float) status->current_postprocessing_cycle) / status->total_postprocessing_cycles;
+				break;
+			case ExportStatus::Tagging:
+				status_text = string_compose (_("Tagging '%3' (timespan %1 of %2)"),
+				                              status->timespan, status->total_timespans, status->timespan_name);
+				break;
+			case ExportStatus::Uploading:
+				status_text = string_compose (_("Uploading '%3' (timespan %1 of %2)"),
+				                              status->timespan, status->total_timespans, status->timespan_name);
+				break;
+			case ExportStatus::Command:
+				status_text = string_compose (_("Running Post Export Command for '%1'"), status->timespan_name);
+				break;
+		}
 	}
 
 	progress_bar.set_text (status_text);
@@ -392,6 +563,7 @@ void
 ExportDialog::add_error (string const & text)
 {
 	export_button->set_sensitive (false);
+	analyze_button->set_sensitive (false);
 
 	if (warn_string.empty()) {
 		warn_string = _("<span color=\"#ffa755\">Error: ") + text + "</span>";
@@ -461,7 +633,7 @@ ExportRegionDialog::init_gui ()
 void
 ExportRegionDialog::init_components ()
 {
-	string loc_id = profile_manager->set_single_range (region.position(), region.position() + region.length(), region.name());
+	string loc_id = profile_manager->set_single_range (region.position_sample(), (region.position() + region.length()).samples(), region.name());
 
 	preset_selector.reset (new ExportPresetSelector ());
 	timespan_selector.reset (new ExportTimespanSelectorSingle (_session, profile_manager, loc_id));
